#ifndef ____AES_H____
#define ____AES_H____

/*  This include is used to find 8 & 32 bit unsigned integer types  */
#include "limits.h"

#if defined(__cplusplus)
extern "C"
{
#endif

#define AES_128     /* define if AES with 128 bit keys is needed    */
//#define AES_192     /* define if AES with 192 bit keys is needed    */
//#define AES_256     /* define if AES with 256 bit keys is needed    */
//#define AES_VAR     /* define if a variable key size is needed      */

#define AES_ENCRYPT /* if support for encryption is needed          */
#define AES_DECRYPT /* if support for decryption is needed          */

#if UCHAR_MAX == 0xff                   /* an unsigned 8 bit type   */
  typedef unsigned char      aes_08t;
#else
# error Please define aes_08t as an 8-bit unsigned integer type in aes.h
#endif

#if UINT_MAX == 4294967295              /* an unsigned 32 bit type  */
  typedef   unsigned int     aes_32t;
#elif ULONG_MAX == 4294967295ul
  typedef   unsigned long    aes_32t;
#else
# error Please define aes_32t as a 32-bit unsigned integer type in aes.h
#endif

#define AES_BLOCK_SIZE  16  /* the AES block size in bytes          */
#define N_COLS           4  /* the number of columns in the state   */

/* The key schedule length is 11, 13 or 15 16-byte blocks for 128,  */
/* 192 or 256-bit keys respectively. That is 176, 208 or 240 bytes  */
/* or 44, 52 or 60 32-bit words. For simplicity this code allocates */
/* the maximum 60 word array for the key schedule for all key sizes */

#if defined( AES_VAR ) || defined( AES_256 )
#define KS_LENGTH       60
#elif defined( AES_192 )
#define KS_LENGTH       52
#else
#define KS_LENGTH       44
#endif

/* the character array 'inf' in the following structures is used	*/
/* to hold AES context information. This AES code uses cx->inf.b[0]	*/
/* to hold the number of rounds multiplied by 16. The other three	*/
/* elements can be used by code that implements additional modes	*/

typedef union 
{
	aes_32t	l;
	aes_08t	b[4];
} aes_inf;

typedef struct
{ 
	aes_32t ks[KS_LENGTH];
	aes_inf inf;
} aes_encrypt_ctx;

typedef struct
{
	aes_32t ks[KS_LENGTH];
	aes_inf inf;
} aes_decrypt_ctx;

/* This routine must be called before first use if non-static       */
/* tables are being used                                            */

void gen_tabs(void);

/* The key length (klen) is input in bytes when it is in the range  */
/* 16 <= klen <= 32 or in bits when in the range 128 <= klen <= 256 */

#if defined( AES_ENCRYPT )

#if defined(AES_128) || defined(AES_VAR)
void aes_encrypt_key128(const unsigned char *key, aes_encrypt_ctx cx[1]);
#endif

#if defined(AES_192) || defined(AES_VAR)
void aes_encrypt_key192(const unsigned char *key, aes_encrypt_ctx cx[1]);
#endif

#if defined(AES_256) || defined(AES_VAR)
void aes_encrypt_key256(const unsigned char *key, aes_encrypt_ctx cx[1]);
#endif

#endif

#if defined( AES_DECRYPT )

#if defined(AES_128) || defined(AES_VAR)
void aes_decrypt_key128(const unsigned char *key, aes_decrypt_ctx cx[1]);
#endif

#if defined(AES_192) || defined(AES_VAR)
void aes_decrypt_key192(const unsigned char *key, aes_decrypt_ctx cx[1]);
#endif

#if defined(AES_256) || defined(AES_VAR)
void aes_decrypt_key256(const unsigned char *key, aes_decrypt_ctx cx[1]);
#endif

#endif

#if defined(__cplusplus)
}
#endif

#endif
